/* Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Kristian Lauszus, TKJ Electronics
Web      :  http://www.tkjelectronics.com
e-mail   :  kristianl@tkjelectronics.com
*/

#ifndef _btd_h_
#define _btd_h_

#include "Usb.h"
#include "confdescparser.h"

/* Bluetooth dongle data taken from descriptors */
#define HCI_BUFFERSIZE         64 // max size for ACL data
#define L2CAP_OUT_BUFFERSIZE   32

// Used in control endpoint header for HCI Commands
#define bmREQ_HCI_OUT USB_SETUP_HOST_TO_DEVICE|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE
#define bmREQ_HCI_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE
// #define bmREQ_HCI_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE
// #define bmREQ_HCI_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_INTERFACE
// #define bmREQ_HCI_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_ENDPOINT
// #define bmREQ_HCI_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_OTHER
// ACL Data-In pipe read request 
#define bmREQ_ACL_IN  USB_SETUP_DEVICE_TO_HOST|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_DEVICE

// Used in control endpoint header for HID Commands
#define bmREQ_HID_OUT USB_SETUP_HOST_TO_DEVICE|USB_SETUP_TYPE_CLASS|USB_SETUP_RECIPIENT_INTERFACE
#define HID_REQUEST_SET_REPORT      0x09

/* Bluetooth HCI states for hci_task() */
#define HCI_INIT_STATE                    0
#define HCI_RESET_STATE                   1
#define HCI_SET_EVTMASK_STATE             2
#define HCI_READ_LOCAL_SUPPORTED_COMMANDS 3
#define HCI_CLEAR_EVENTS_STATE            4
#define HCI_SET_EVENTS_STATE              5     
#define HCI_WRITE_AUTHENTICATION_ENABLE   6
#define HCI_CLASS_STATE                   7
#define HCI_BDADDR_STATE                  8 
#define HCI_LOCAL_VERSION_STATE           9
#define HCI_SET_NAME_STATE                10
#define HCI_CHECK_WII_SERVICE             11 // These three states are only used if it should pair and connect to a Wii controller
#define HCI_INQUIRY_STATE                 12
#define HCI_CONNECT_WII_STATE             13
#define HCI_CONNECTED_WII_STATE           14
#define HCI_SCANNING_STATE                15
#define HCI_CONNECT_IN_STATE              16
#define HCI_REMOTE_NAME_STATE             17
#define HCI_CONNECTED_STATE               18
#define HCI_DISABLE_SCAN_STATE            19 
#define HCI_DONE_STATE                    20
#define HCI_DISCONNECT_STATE              21 

/* HCI event flags*/
#define HCI_FLAG_CMD_COMPLETE           0x01
#define HCI_FLAG_CONN_COMPLETE          0x02
#define HCI_FLAG_DISCONN_COMPLETE       0x04
#define HCI_FLAG_REMOTE_NAME_COMPLETE   0x08
#define HCI_FLAG_INCOMING_REQUEST       0x10
#define HCI_FLAG_READ_BDADDR            0x20
#define HCI_FLAG_READ_VERSION           0x40
#define HCI_FLAG_NONEMPTY_INQUIRY_RESULT              0x80
#define HCI_FLAG_CONNECT_EVENT          0x100

/*Macros for HCI event flag tests */
#define hci_cmd_complete (hci_event_flag & HCI_FLAG_CMD_COMPLETE)
#define hci_connect_complete (hci_event_flag & HCI_FLAG_CONN_COMPLETE)
#define hci_disconnect_complete (hci_event_flag & HCI_FLAG_DISCONN_COMPLETE)
#define hci_remote_name_complete (hci_event_flag & HCI_FLAG_REMOTE_NAME_COMPLETE)
#define hci_incoming_connect_request (hci_event_flag & HCI_FLAG_INCOMING_REQUEST)
#define hci_read_bdaddr_complete (hci_event_flag & HCI_FLAG_READ_BDADDR)
#define hci_read_version_complete (hci_event_flag & HCI_FLAG_READ_VERSION)
#define hci_wii_found (hci_event_flag & HCI_FLAG_NONEMPTY_INQUIRY_RESULT)
#define hci_connect_event (hci_event_flag & HCI_FLAG_CONNECT_EVENT)

/* HCI Events managed */
#define EV_INQUIRY_COMPLETE                             0x01
#define EV_INQUIRY_RESULT                               0x02
#define EV_CONNECT_COMPLETE                             0x03
#define EV_INCOMING_CONNECT                             0x04
#define EV_DISCONNECT_COMPLETE                          0x05
#define EV_AUTHENTICATION_COMPLETE                      0x06
#define EV_REMOTE_NAME_COMPLETE                         0x07
#define EV_ENCRYPTION_CHANGE                            0x08
#define EV_CHANGE_CONNECTION_LINK                       0x09
#define EV_ROLE_CHANGED                                 0x12
#define EV_NUM_COMPLETE_PKT                             0x13
#define EV_PIN_CODE_REQUEST                             0x16
#define EV_LINK_KEY_REQUEST                             0x17
#define EV_LINK_KEY_NOTIFICATION                        0x18
#define EV_DATA_BUFFER_OVERFLOW                         0x1A
#define EV_MAX_SLOTS_CHANGE                             0x1B
#define EV_READ_REMOTE_VERSION_INFORMATION_COMPLETE     0x0C
#define EV_QOS_SETUP_COMPLETE                           0x0D
#define EV_COMMAND_COMPLETE                             0x0E
#define EV_COMMAND_STATUS                               0x0F
#define EV_LOOPBACK_COMMAND                             0x19
#define EV_PAGE_SCAN_REP_MODE                           0x20

/* Event Filter constants (hci_set_event_filter OCF=0x05, OGF=0x03) */
#define EV_FILTER_TYPE_CLEAR_ALL          0x00
#define EV_FILTER_TYPE_INQUIRY_RESULT     0x01
#define EV_FILTER_TYPE_CONNECTION_SETUP   0x02

#define EV_FILTER_COND_INQ_ACCEPT_ALL       0x00
#define EV_FILTER_COND_INQ_CLASS_RESPONDER  0x01
#define EV_FILTER_COND_INQ_ADDR_RESPONDER   0x02
#define EV_FILTER_COND_INQ_CLASS_ANY        0x000000
#define EV_FILTER_COND_INQ_CLASS_ANYMASK    0x000000
#define EV_FILTER_COND_INQ_ADDR_ANY         0x000000000000

#define EV_FILTER_COND_CONN_ACCEPT_ALL       0x00
#define EV_FILTER_COND_CONN_CLASS_MATCHED    0x01
#define EV_FILTER_COND_CONN_ADDR_MATCHED     0x02

#define EV_FILTER_COND_CONN_ALL_NOAUTO          0x01
#define EV_FILTER_COND_CONN_ALL_NO_ROLESWITCH   0x02
#define EV_FILTER_COND_CONN_ALL_WITH_ROLESWITCH 0x03
#define EV_FILTER_COND_CONN_CLASS_ANY           0x000000
#define EV_FILTER_COND_CONN_CLASS_ANYMASK       0x000000
#define EV_FILTER_COND_CONN_ADDR_ANY            0x000000000000

/* L2CAP signaling commands */
#define L2CAP_CMD_COMMAND_REJECT        0x01
#define L2CAP_CMD_CONNECTION_REQUEST    0x02
#define L2CAP_CMD_CONNECTION_RESPONSE   0x03
#define L2CAP_CMD_CONFIG_REQUEST        0x04
#define L2CAP_CMD_CONFIG_RESPONSE       0x05
#define L2CAP_CMD_DISCONNECT_REQUEST    0x06
#define L2CAP_CMD_DISCONNECT_RESPONSE   0x07
#define L2CAP_CMD_INFORMATION_REQUEST   0x0A
#define L2CAP_CMD_INFORMATION_RESPONSE  0x0B

// Used For Connection Response - Remember to Include High Byte
#define PENDING     0x01
#define SUCCESSFUL  0x00

/* Bluetooth L2CAP PSM - see http://www.bluetooth.org/Technical/AssignedNumbers/logical_link.htm */
#define SDP_PSM         0x01 // Service Discovery Protocol PSM Value
#define RFCOMM_PSM      0x03 // RFCOMM PSM Value
#define HID_CTRL_PSM    0x11 // HID_Control PSM Value
#define HID_INTR_PSM    0x13 // HID_Interrupt PSM Value

// Used to determine if it is a Bluetooth dongle
#define WI_SUBCLASS_RF      0x01 // RF Controller
#define WI_PROTOCOL_BT      0x01 // Bluetooth Programming Interface

#define RESET_DEFERRAL_LOOPS 100 // Loop this many times before trying to send the hci reset command
#define BTD_MAX_ENDPOINTS   4
#define BTD_NUMSERVICES     4 // Max number of Bluetooth services - if you need more than four simply increase this number
#define BTD_ENDPOINT_POLLING_INTERVAL 1 // mSec, default 1 millisecond

// DISCARDABLE CONSTANTS
/* Host error result codes, the 4 LSB's in the HRSL register */
#define hrSUCCESS   0x00
#define hrBUSY      0x01
#define hrBADREQ    0x02
#define hrUNDEF     0x03
#define hrNAK       0x04
#define hrSTALL     0x05
#define hrTOGERR    0x06
#define hrWRONGPID  0x07
#define hrBADBC     0x08
#define hrPIDERR    0x09
#define hrPKTERR    0x0A
#define hrCRCERR    0x0B
#define hrKERR      0x0C
#define hrJERR      0x0D
#define hrTIMEOUT   0x0E
#define hrBABBLE    0x0F

/* Bluetooth L2CAP states for SDP_task() */
#define L2CAP_SDP_WAIT                  0
#define L2CAP_SDP_REQUEST               1
#define L2CAP_SDP_SUCCESS               2
#define L2CAP_SDP_DONE                  3
#define L2CAP_DISCONNECT_RESPONSE       4

/* Bluetooth L2CAP states for RFCOMM_task() */
#define L2CAP_RFCOMM_WAIT               0
#define L2CAP_RFCOMM_REQUEST            1
#define L2CAP_RFCOMM_SUCCESS            2
#define L2CAP_RFCOMM_DONE               3

/* L2CAP event flags */
#define L2CAP_FLAG_CONNECTION_SDP_REQUEST       0x001
#define L2CAP_FLAG_CONNECTION_RFCOMM_REQUEST    0x002
#define L2CAP_FLAG_CONFIG_SDP_REQUEST           0x004
#define L2CAP_FLAG_CONFIG_RFCOMM_REQUEST        0x008
#define L2CAP_FLAG_CONFIG_SDP_SUCCESS           0x010
#define L2CAP_FLAG_CONFIG_RFCOMM_SUCCESS        0x020
#define L2CAP_FLAG_DISCONNECT_SDP_REQUEST       0x040
#define L2CAP_FLAG_DISCONNECT_RFCOMM_REQUEST    0x080
#define L2CAP_FLAG_DISCONNECT_RESPONSE          0x100

/* Macros for L2CAP event flag tests */
#define l2cap_connection_request_sdp_flag (l2cap_event_flag & L2CAP_FLAG_CONNECTION_SDP_REQUEST)
#define l2cap_connection_request_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_CONNECTION_RFCOMM_REQUEST)
#define l2cap_config_request_sdp_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_SDP_REQUEST)
#define l2cap_config_request_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_RFCOMM_REQUEST)
#define l2cap_config_success_sdp_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_SDP_SUCCESS)
#define l2cap_config_success_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_CONFIG_RFCOMM_SUCCESS)
#define l2cap_disconnect_request_sdp_flag (l2cap_event_flag & L2CAP_FLAG_DISCONNECT_SDP_REQUEST)
#define l2cap_disconnect_request_rfcomm_flag (l2cap_event_flag & L2CAP_FLAG_DISCONNECT_RFCOMM_REQUEST)
#define l2cap_disconnect_response_flag (l2cap_event_flag & L2CAP_FLAG_DISCONNECT_RESPONSE)

/* Used for SDP */
#define SDP_SERVICE_SEARCH_ATTRIBUTE_REQUEST_PDU    0x06 // See the RFCOMM specs
#define SDP_SERVICE_SEARCH_ATTRIBUTE_RESPONSE_PDU   0x07 // See the RFCOMM specs
#define SERIALPORT_UUID     0x1101 // See http://www.bluetooth.org/Technical/AssignedNumbers/service_discovery.htm
#define L2CAP_UUID          0x0100

/* Used for RFCOMM */
#define RFCOMM_SABM     0x2F
#define RFCOMM_UA       0x63
#define RFCOMM_UIH      0xEF
//#define RFCOMM_DM       0x0F
#define RFCOMM_DISC     0x43

#define extendAddress   0x01 // Always 1

// Multiplexer message types
#define BT_RFCOMM_PN_CMD     0x83
#define BT_RFCOMM_PN_RSP     0x81
#define BT_RFCOMM_MSC_CMD    0xE3
#define BT_RFCOMM_MSC_RSP    0xE1
#define BT_RFCOMM_RPN_CMD    0x93
#define BT_RFCOMM_RPN_RSP    0x91
/*
#define BT_RFCOMM_TEST_CMD   0x23
#define BT_RFCOMM_TEST_RSP   0x21
#define BT_RFCOMM_FCON_CMD   0xA3
#define BT_RFCOMM_FCON_RSP   0xA1
#define BT_RFCOMM_FCOFF_CMD  0x63
#define BT_RFCOMM_FCOFF_RSP  0x61
#define BT_RFCOMM_RLS_CMD    0x53
#define BT_RFCOMM_RLS_RSP    0x51
#define BT_RFCOMM_NSC_RSP    0x11
*/

#define USB_NORESP_LIMIT    10
#define noresp_limit        ((USB_NORESP_LIMIT > 0 && hci_counter++ > USB_NORESP_LIMIT) ? ((hci_counter = -1), 1) : 0)

/** All Bluetooth services should include this class. */
class BluetoothService {
  public:
    /**
     * Used to pass acldata to the Bluetooth service.
     * @param ACLData Pointer to the incoming acldata.
     */
    virtual void ACLData(uint8_t* ACLData);
    /** Used to run the different state machines in the Bluetooth service. */
    virtual void Run(EpInfo &);
    /** Used to reset the Bluetooth service. */
    virtual void Reset();
    /** Used to disconnect both the L2CAP Channel and the HCI Connection for the Bluetooth service. */
    virtual void disconnect();
};

class BTD;

typedef void (BTD::*HCI_State_Setter)();

/**
 * The Bluetooth Dongle class will take care of all the USBHost communication
 * and then pass the data to the BluetoothService classes.
 */
class BTD : public USBDeviceConfig, public UsbConfigXtracter {/*{{{*/

  friend class SPP;

  public:
    /**
     * Constructor for the BTD class.
     * @param  p   Pointer to USBHost class instance.
     */
    BTD(USBHost *p, const char* name = "Arduino", const char* pin = "1234");

    virtual void setName(const char *name);

    virtual void setPin(const char *pin);

    /** @name USBDeviceConfig implementation */
    /**
     * Address assignment and basic initilization is done here.
     * @param  parent   Hub number.
     * @param  port     Port number on the hub.
     * @param  lowspeed Speed of the device.
     * @return          0 on success.
     */
    virtual uint32_t ConfigureDevice(uint32_t parent, uint32_t port, uint32_t lowspeed);
    /**
     * Initialize the Bluetooth dongle.
     * @param  parent   Hub number.
     * @param  port     Port number on the hub.
     * @param  lowspeed Speed of the device.
     * @return          0 on success.
     */
    virtual uint32_t Init(uint32_t parent, uint32_t port, uint32_t lowspeed);
    /**
     * Release the USB device.
     * @return 0 on success.
     */
    virtual uint32_t Release();
    /**
     * Poll the USB Input endpoins and run the state machines.
     * @return 0 on success.
     */
    virtual uint32_t Poll();

    /**
     * Get the device address.
     * @return The device address.
     */
    virtual uint32_t GetAddress() {
      return bAddress;
    };

    /**
     * Used to check if the dongle has been initialized.
     * @return True if it's ready.
     */
    virtual bool isReady() {
      return bPollEnable;
    };
    /**
     * Used by the USB core to check what this driver support.
     * @param  klass The device's USB class.
     * @return       Returns true if the device's USB class matches this driver.
     */
    virtual boolean DEVCLASSOK(uint8_t klass) { return (klass == USB_CLASS_WIRELESS_CTRL); }

    /**
     * Used by the USB core to check what this driver support.
     * Used to set the Bluetooth address into the PS3 controllers.
     * @param  vid The device's VID.
     * @param  pid The device's PID.
     * @return     Returns true if the device's VID and PID matches this driver.
     */
    virtual boolean VIDPIDOK(uint16_t vid, uint16_t pid) {
      return true; 
    };

    /** @name UsbConfigXtracter implementation */
    /**
     * UsbConfigXtracter implementation, used to extract endpoint information.
     * @param conf  Configuration value.
     * @param iface Interface number.
     * @param alt   Alternate setting.
     * @param proto Interface Protocol.
     * @param ep    Endpoint Descriptor.
     */
    virtual void EndpointXtract(uint32_t conf, uint32_t iface, uint32_t alt, uint32_t proto, const USB_ENDPOINT_DESCRIPTOR *ep);

    /** Disconnects both the L2CAP Channel and the HCI Connection for all Bluetooth services. */
    void disconnect() {
      for(uint8_t i = 0; i < BTD_NUMSERVICES; i++)
        if(!(NULL == btService[i])) {
          btService[i]->disconnect();
          btService[i] = NULL;
        }
    };

    /**
     * Register bluetooth dongle members/services.
     * @param  pService Pointer to BluetoothService class instance.
     * @return          The serice ID on succes or -1 on fail.
     */
    int8_t registerServiceClass(BluetoothService *pService) {
      for(uint8_t i = 0; i < BTD_NUMSERVICES; i++) {
        if(NULL == btService[i]) {
          btService[i] = pService;
          return i; // Return ID
        }
      }
      return -1; // ErrorregisterServiceClass
    };

    /** @name HCI Commands */
    /**
     * Used to send a HCI Command.
     * @param data   Data to send.
     * @param nbytes Number of bytes to send.
     */
    void HCI_Command(uint8_t* data, uint16_t send_length, uint16_t read_length = 1);
    /** Read supported commands */
    void hci_read_local_supported_commands();
    /** Set HCI default event mask */
    void hci_set_default_event_mask();
    /** Clear/Set Event Filter (OGF = 3, OCF = 5) */
    void hci_clear_event_filter();
    void hci_set_event_filter(uint8_t filter_type, uint8_t filter_cond = 0, uint8_t param_len = 0, uint8_t * filter_params = NULL);
    void hci_set_default_event_filter();
    /** Reset the Bluetooth dongle. */
    void hci_reset();
    /** Read the Bluetooth address of the dongle. */
    void hci_read_bdaddr();
    /** Read the HCI Version of the Bluetooth dongle. */
    void hci_read_local_version_information();
    /**
     * Set the local name of the Bluetooth dongle.
     * @param name Desired name.
     */
    void hci_set_local_name(const char* name);
    /** Enable visibility to other Bluetooth devices. */
    void hci_write_scan_enable();
    /** Disable visibility to other Bluetooth devices. */
    void hci_write_scan_disable();
    /** Read the remote devices name. */
    void hci_remote_name();
    /** Accept the connection with the Bluetooth device. */
    void hci_accept_connection();
    /**
     * Disconnect the HCI connection.
     * @param handle The HCI Handle for the connection.
     */
    void hci_disconnect(uint16_t handle);
    /**
     * Respond with the pin for the connection.
     * The pin is automatically set for the Wii library,
     * but can be customized for the SPP library.
     */
    void hci_pin_code_request_reply();
    /** Respons when no pin was set. */
    void hci_pin_code_negative_request_reply();
    /**
     * Command is used to reply to a Link Key Request event from the BR/EDR Controller
     * if the Host does not have a stored Link Key for the connection.
     */
    void hci_link_key_request_negative_reply();
    /** Set/clear authentication enable flag. */
    void hci_write_authentication_enable(bool);
    /** Used to try to authenticate with the remote device. */
    void hci_authentication_request();
    /** Start a HCI inquiry. */
    void hci_inquiry();
    /** Cancel a HCI inquiry. */
    void hci_inquiry_cancel();
    /** Connect to a device. */
    void hci_connect();
    /** Used to a set the class of the device. */
    void hci_write_class_of_device();

    /** @name L2CAP Commands */
    /**
     * Used to send L2CAP Commands.
     * @param handle      HCI Handle.
     * @param data        Data to send.
     * @param nbytes      Number of bytes to send.
     * @param channelLow,channelHigh  Low and high byte of channel to send to.
     * If argument is omitted then the Standard L2CAP header: Channel ID (0x01) for ACL-U will be used.
     */
    void L2CAP_Command(uint16_t handle, uint8_t* data, uint8_t nbytes, uint8_t channelLow = 0x01, uint8_t channelHigh = 0x00);
    /**
     * L2CAP Connection Request.
     * @param handle HCI handle.
     * @param rxid   Identifier.
     * @param scid   Source Channel Identifier.
     * @param psm    Protocol/Service Multiplexer - see: https://www.bluetooth.org/Technical/AssignedNumbers/logical_link.htm.
     */
    void l2cap_connection_request(uint16_t handle, uint8_t rxid, uint8_t* scid, uint16_t psm);
    /**
     * L2CAP Connection Response.
     * @param handle HCI handle.
     * @param rxid   Identifier.
     * @param dcid   Destination Channel Identifier.
     * @param scid   Source Channel Identifier.
     * @param result Result - First send ::PENDING and then ::SUCCESSFUL.
     */
    void l2cap_connection_response(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint8_t result);
    /**
     * L2CAP Config Request.
     * @param handle HCI Handle.
     * @param rxid   Identifier.
     * @param dcid   Destination Channel Identifier.
     */
    void l2cap_config_request(uint16_t handle, uint8_t rxid, uint8_t* dcid);
    /**
     * L2CAP Config Response.
     * @param handle HCI Handle.
     * @param rxid   Identifier.
     * @param scid   Source Channel Identifier.
     */
    void l2cap_config_response(uint16_t handle, uint8_t rxid, uint8_t* scid);
    /**
     * L2CAP Disconnection Request.
     * @param handle HCI Handle.
     * @param rxid   Identifier.
     * @param dcid   Device Channel Identifier.
     * @param scid   Source Channel Identifier.
     */
    void l2cap_disconnection_request(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid);
    /**
     * L2CAP Disconnection Response.
     * @param handle HCI Handle.
     * @param rxid   Identifier.
     * @param dcid   Device Channel Identifier.
     * @param scid   Source Channel Identifier.
     */
    void l2cap_disconnection_response(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid);
    /**
     * L2CAP Information Response.
     * @param handle       HCI Handle.
     * @param rxid         Identifier.
     * @param infoTypeLow,infoTypeHigh  Infotype.
     */
    void l2cap_information_response(uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh);

    /** Use this to see if it is waiting for a incoming connection. */
    bool connectionWait;
    /** This is used by the service to know when to store the device information. */
    bool l2capConnectionClaimed;
    /** This is used by the SPP library to claim the current SDP incoming request. */
    bool sdpConnectionClaimed;
    /** This is used by the SPP library to claim the current RFCOMM incoming request. */
    bool rfcommConnectionClaimed;

    /** The name you wish to make the dongle show up as. It is set automatically by the SPP library. */
    char btdName[250];
    /** The pin you wish to make the dongle use for authentication. It is set automatically by the SPP library. */
    char btdPin[17]; // NUL-terminated, up to 128 bits long

    /** The Bluetooth dongle's Bluetooth address. */
    uint8_t my_bdaddr[6];
    /** HCI handle for the last connection. */
    uint16_t hci_handle;
    /** Last incoming device's Bluetooth address. */
    uint8_t disc_bdaddr[6];
    /** First 30 chars of last remote name. */
    uint8_t remote_name[30];
    /**
     * The supported HCI Version read from the Bluetooth dongle.
     * Used by the PS3BT library to check the HCI Version of the Bluetooth dongle,
     * it should be at least 3 to work properly with the library.
     */
    uint8_t hci_version;

    /**
     * Read the poll interval taken from the endpoint descriptors.
     * @return The poll interval in ms.
     */
    uint8_t readPollInterval() {
      return pollInterval;
    };

  protected:
    /** Pointer to USB class instance. */
    USBHost *pUsb;
    /** Pointer to last-invoked HCI state handler */
    HCI_State_Setter hci_last_invoked;
    /** Device address. */
    uint32_t bAddress;
    /** Number of configurations */
    uint8_t bNumConfigurations;
    /** Maximum packet size */
    uint8_t maxPktSize;
    /** Endpoint info structure. */
    EpInfo epInfo[BTD_MAX_ENDPOINTS];

    /** Configuration number. */
    uint32_t bConfNum;
    /** Total number of endpoints in the configuration. */
    uint32_t bNumEP;
    /** Next poll time based on poll interval taken from the USB descriptor. */
    uint32_t qNextPollTime;
    /** SAM internal tick count */
    uint32_t ticks;

    /** Bluetooth dongle control endpoint. */
    static const uint32_t BTD_CONTROL_PIPE;
    /** HCI event endpoint index. */
    static const uint32_t BTD_EVENT_PIPE;
    /** ACL In endpoint index. */
    static const uint32_t BTD_DATAIN_PIPE;
    /** ACL Out endpoint index. */
    static const uint32_t BTD_DATAOUT_PIPE;

    /**
     * Used to print the USB Endpoint Descriptor.
     * @param ep_ptr Pointer to USB Endpoint Descriptor.
     */
    void PrintEndpointDescriptor(const USB_ENDPOINT_DESCRIPTOR* ep_ptr);

  private:

    void clearAllVariables(); // Set all variables, endpoint structs etc. to default values

    BluetoothService* btService[BTD_NUMSERVICES];
    uint16_t PID, VID; // PID and VID of device connected

    bool bPollEnable;
    bool bAuthState;
    uint8_t pollInterval;

    /* Variables used by high level HCI task */
    uint8_t hci_state; //current state of bluetooth hci connection
    uint32_t hci_counter; // counter used for bluetooth hci reset loops
    uint32_t hci_num_reset_loops; // this value indicate how many times it should read before trying to reset
    uint16_t hci_event_flag; // hci flags of received bluetooth events
    uint8_t inquiry_counter;

    uint8_t hcibuf[HCI_BUFFERSIZE]; //General purpose buffer for hci data
    uint8_t l2capoutbuf[L2CAP_OUT_BUFFERSIZE]; //General purpose buffer for l2cap out data

    /* State machines */
    void HCI_event_task(); // Poll the HCI event pipe
    void HCI_task(); // HCI state machine

    /* Used to set the Bluetooth Address internally to the PS3 Controllers */
    void setBdaddr(uint8_t* BDADDR);
    void setMoveBdaddr(uint8_t* BDADDR);
};/*}}}*/


#endif
